/** svgtomap.d
 * 
 * Command-line application for turning SVG files into Mayhem Intergalactic map
 * files. This allows Inkscape (and other SVG editors) to be used as map editors
 * for Mayhem Intergalactic. Primitive, but functional.
 * 
 * In Inkscape, draw circles for planets and use the straight-line tool for rifts.
 * 
 * Version history:
 * 03/03/2009	v1.0	Initial release
 * 
 * Written in the D programming language (http://www.digitalmars.com/d/)
 * Uses the XML library for D written by Miguel Ferreira Simoes.
 * 
 * To compile, open up a command line and type:
 * > dmd svgtomap.d xml\XmlNode.d xml\XmlStream.d xml\XmlEntities.d xml\XmlException.d -ofsvgtomap.exe
 * 
 * (You must first install the official D compiler, dmd.)
 * 
 * Licence notice follows.
 * 
 * Copyright (c) 2009 Inventive Dingo
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 *     1. The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 * 
 *     2. Altered source versions must be plainly marked as such, and must not be
 *     misrepresented as being the original software.
 * 
 *     3. This notice may not be removed or altered from any source
 *     distribution.
 *
 */

module svgtomap;

import xml.XmlNode;
static import std.file, std.math;
import std.path, std.stdio, std.stream, std.string;


/// "Junk" characters are to be ignored when parsing path data / numbers
bool isJunk(char c) {
	if (c!='-' && find(letters, c)==-1 && find(digits, c)==-1) return true;
	else return false;
}


/// Parse a number from string, starting at index ch, and setting ch past the end of the number
/// before returning.
double parseNumber(in char[] string, inout int ch) {	
	while (isJunk(string[ch])) ch++;
	
	// Negative?
	bool negative = false;
	if (string[ch] == '-') {
		negative = true;
		ch++;
	}
	
	// Get the integer part of the number
	double number=0;
	while (ch<string.length && find(digits, string[ch])!=-1) {
		number *= 10;
		number += atoi(string[ch..ch+1]);
		ch++;
	}
	
	// Get a decimal point and following digits
	if (ch<string.length && string[ch]=='.') {
		ch++;
		double decimalpart = 0;
		int divisor = 1;
		while (ch<string.length && find(digits, string[ch])!=-1) {
			divisor *= 10;
			decimalpart *= 10;
			decimalpart += atoi(string[ch..ch+1]);
			ch++;
		}
		
		number += decimalpart / divisor;
	}
	
	return negative? -number : number;
}


class Matrix {
	const int SIZE = 3;
	double[SIZE][SIZE] data;
	
	this() {}
	this(double d1, double d2, double d3, double d4, double d5, double d6, double d7, double d8, double d9) {
		data[0][0] = d1;
		data[0][1] = d2;
		data[0][2] = d3;
		data[1][0] = d4;
		data[1][1] = d5;
		data[1][2] = d6;
		data[2][0] = d7;
		data[2][1] = d8;
		data[2][2] = d9;
	}
	
	/// Parse an SVG "transform" attribute and return a new matrix corresponding to it
	static Matrix fromSVG(char[] transform) {
		transform = strip(transform);
		
		bool startsWith(char[] str, char[] prefix) {
			if (str.length < prefix.length) return false;
			
			for (int i=0; i<prefix.length; i++) {
				if (prefix[i] != str[i]) return false;
			}
			return true;
		}
		
		// Find first opening bracket
		int firstBracket = -1;
		for (int i=0; i<transform.length; i++) {
			if (transform[i]=='(') {
				firstBracket = i;
				break;
			}
		}
		// Give up if data nonsensical or blank
		if (firstBracket==-1) return makeIdentity();
		
		// Find first closing bracket
		int firstCloseBracket = -1;
		for (int i=0; i<transform.length; i++) {
			if (transform[i]==')') {
				firstCloseBracket = i;
				break;
			}
		}
		// Give up if data nonsensical
		if (firstCloseBracket==-1 || firstCloseBracket < firstBracket) return makeIdentity();
		
		char[] str = transform[firstBracket+1..firstCloseBracket];
		int index = 0;
		Matrix m = null;
		if (startsWith(transform, "matrix")) {
			double n1 = parseNumber(str, index);
			double n2 = parseNumber(str, index);
			double n3 = parseNumber(str, index);
			double n4 = parseNumber(str, index);
			double n5 = parseNumber(str, index);
			double n6 = parseNumber(str, index);
			m = new Matrix(n1, n3, n5, n2, n4, n6, 0, 0, 1);
			
		} else if (startsWith(transform, "translate")) {
			double tx = parseNumber(str, index);
			double ty = parseNumber(str, index);
			m = new Matrix(1, 0, tx, 0, 1, ty, 0, 0, 1);
			
		} else {
			writefln("WARNING: Transform type %s not implemented.\n", transform[0..firstBracket]);
			m = makeIdentity();
		}
		
		return m * fromSVG(transform[firstCloseBracket+1..$]);
	}
	
	static Matrix makeIdentity() {
		auto m = new Matrix();
		for (int i=0; i<SIZE; i++) {
			for (int j=0; j<SIZE; j++) {
				if (i==j) m.data[i][j] = 1;
				else m.data[i][j] = 0;
			}
		}
		return m;
	}
	
	Matrix opMul(Matrix m) {
		auto result = new Matrix();
		
		// Iterate rows of [this], cols of [m]
		for (int row=0; row<SIZE; row++) {
			// Iterate across the row in [result]
			for (int resultcol=0; resultcol<SIZE; resultcol++) {
				result.data[row][resultcol] = 0;
				// Iterate columns of [this], rows of [m]
				for (int col=0; col<SIZE; col++) {
					result.data[row][resultcol] += this.data[row][col]*m.data[col][resultcol];
				}
			}
		}
		return result;
	}
	
	void applyToVector(inout double x, inout double y) {
		double newX = data[0][0]*x + data[0][1]*y + data[0][2];
		double newY = data[1][0]*x + data[1][1]*y + data[1][2];
		x = newX;
		y = newY;
	}
	
	bool opEquals(Matrix m) {
		for (int i=0; i<SIZE; i++) {
			for (int j=0; j<SIZE; j++) {
				if (data[i][j] != m.data[i][j]) {
					writefln("At %d,%d: this has %f, m has %f", i, j, data[i][j], m.data[i][j]);
					return false;
				}
			}
		}
		return true;
	}
	
	char[] toString() {
		char[] result = "";
		for (int i=0; i<SIZE; i++) {
			result ~= "[";
			for (int j=0; j<SIZE; j++) {
				result ~= .toString(data[i][j])~" ";
			}
			result ~= "] ";
		}
		return result;
	}
}


unittest {
	auto identity = new Matrix(1,0,0,  0,1,0,  0,0,1);
	auto identitysq = identity*identity;
	assert(identitysq == identity, "identity is "~identity.toString()~", squared it's "~identitysq.toString());
	
	double x = 5, y = 6;
	identity.applyToVector(x, y);
	assert(x==5 && y==6);
	
	auto test = new Matrix(1,2,3,4,5,6,7,8,9);
	assert(identity*test == test, "test is "~test.toString()~"; * identity it's "~(identity*test).toString());
	
	auto testsq = new Matrix(30, 36, 42, 66, 81, 96, 102, 126, 150);
	assert(testsq == test*test, "test squared should be "~testsq.toString()~", but it's coming out as "~(test*test).toString());
}


/// Variable scale factor
float scaleFactor=1;

int scale(double x) {
	double f = 1.5*scaleFactor*x;
	return cast(int)std.math.round(f);
}


/// Interpret path data as planets and rifts
char[] parsePathData(char[] pathData, Matrix transform, char[] id, char[] label) {
	double penX=0, penY=0; // Current pen location
	double subpathStartX=0, subpathStartY=0; // Start of current subpath
	
	//writefln("Parsing path data: %s (matrix %s)", pathData, transform);
	
	// For averaging the coords given for each planet
	int planetCoordsCount=0;
	double planetCoordsTotalX=0, planetCoordsTotalY=0;
	
	int ch=0; // Index of next character to parse
	bool parsingPlanet = false; // True if we're definitely parsing a planet, not a rift
	char[] iniData;
	
	//writefln("%s", pathData);
	
	bool numberIsNext() {
		//writef("Scanning for number. Finding junk characters: {");
		int localch = ch;
		while (localch < pathData.length && isJunk(pathData[localch])) {
			//writef("%s", pathData[localch..localch+1]);
			localch++;
		}
		//writefln("}");
		
		if (localch >= pathData.length) return false;
		
		if (find(digits, pathData[localch])==-1) {
			//writefln("%s does NOT start with digit: %s, therefore number is NOT next", pathData[localch..$], [pathData[localch]]);
			return false;
		} else {
			//writefln("%s starts with digit: %s, therefore number is next", pathData[localch..$], [pathData[localch]]);
			return true;
		}
	}
	
	double getNumber() {
		assert(numberIsNext());
		
		double number = parseNumber(pathData, ch);
		
		//writefln("getNumber() returning %f", number);
		
		return number;
	}
	
	double getXCoord() {
		if (toupper( [pathData[ch-1]] )[0] == pathData[ch-1]) {
			// Absolute
			return getNumber();
		} else {
			// Relative
			return penX + getNumber();
		}
	}
	
	double getYCoord() {
		if (toupper( [pathData[ch-1]] )[0] == pathData[ch-1]) {
			// Absolute
			return getNumber();
		} else {
			// Relative
			return penY + getNumber();
		}
	}
	
	int nextRiftIndex = 1;
	void endRift() {
		if (parsingPlanet) return;
		if (nextRiftIndex != 1) {
			nextRiftIndex = 1;
		}
	}
	
	void drawRiftTo(in double penSpaceNewX, in double penSpaceNewY) {
		if (parsingPlanet) return;
		
		double newX = penSpaceNewX, newY = penSpaceNewY;
		transform.applyToVector(newX, newY);
		double real_penX = penX, real_penY = penY;
		transform.applyToVector(real_penX, real_penY);
		
		if (real_penX != newX || real_penY != newY) {
			auto rift = new Rift(new Point2D(scale(real_penX), scale(real_penY)), new Point2D(scale(newX), scale(newY)));
			global_rifts ~= rift;
			
			if (nextRiftIndex==1) {
				iniData ~= "\n[rift]";
				if (id.length)    iniData ~= "\n; ID: "~id;
				if (label.length) iniData ~= "\n; Label: "~label;
				iniData ~= "\nx1="~.toString(scale(real_penX))
					      ~"\ny1="~.toString(scale(real_penY))~"\n\n";
			} else {
				iniData ~= "\n";
			}
			nextRiftIndex++;
			
			iniData ~= "x"~.toString(nextRiftIndex)~"="~.toString(scale(newX))~"\n";
			iniData ~= "y"~.toString(nextRiftIndex)~"="~.toString(scale(newY))~"\n";
		}
		
		penX = penSpaceNewX;
		penY = penSpaceNewY;
		
		// Minor hack: Make all rifts contain only one line
		endRift();
		
	}
	
	void endPlanet() {
		if (planetCoordsCount) {
			double unscaled_x = planetCoordsTotalX/planetCoordsCount;
			double unscaled_y = planetCoordsTotalY/planetCoordsCount;
			//iniData ~= "\n\n; Raw coords: "~.toString(unscaled_x)~", "~.toString(unscaled_y)~"\n"; // DEBUG
			//iniData ~= "; Matrix: "~transform.toString()~"\n"; // DEBUG
			transform.applyToVector(unscaled_x, unscaled_y);
			//iniData ~= "; Transformed coords: "~.toString(unscaled_x)~", "~.toString(unscaled_y)~"\n"; // DEBUG
			auto x = scale(unscaled_x);
			auto y = scale(unscaled_y);
			//iniData ~= "; Scaled coords: "~.toString(x)~", "~.toString(y)~"\n"; // DEBUG
			
			auto planet = new Planet(x, y);
			global_planets ~= planet;
			//iniData ~= "; "~planet.toString(); // For debugging
			
			iniData ~= "\n[planet]";
			if (id.length)    iniData ~= "\n; ID: "~id;
			if (label.length) iniData ~= "\nname="~label;
			iniData ~= "\nx="~.toString(x)
			         ~ "\ny="~.toString(y)~"\n";
		}
		planetCoordsCount=0;
		planetCoordsTotalX=0;
		planetCoordsTotalY=0;
		parsingPlanet = false;
	}
	
	void handleLineWithSkippedArgs(char instruction, int skip) {
		do {
			// Skip some arguments
			for (int i=0; i<skip; i++) getNumber();
			
			// Draw a rift from the old pen position to the new pen position
			auto newX = penX;
			auto newY = penY;
			if (instruction!='V') newX = getXCoord();
			if (instruction!='H') newY = getYCoord();
			drawRiftTo(newX, newY);
		} while(numberIsNext());
	}
	
	while(ch < pathData.length) {
		// Find a command
		ch++;
		char c = toupper( [pathData[ch-1]] )[0];
		switch (c) {
			case 'M':
				endRift();
				endPlanet();
				do {
					subpathStartX = penX = getXCoord();
					subpathStartY = penY = getYCoord();
				} while(numberIsNext());
				break;
			case 'Z':
				// Draw a rift from the pen to the beginning of this subpath.
				// TODO: Or, finish a planet.
				drawRiftTo(subpathStartX, subpathStartY);
				break;
			case 'L': // Ordinary line
			case 'H': // Horizontal line
			case 'V': // Vertical line
			case 'T': // Shorthand bezier curve (we ignore the curve and simply take the destination)
				handleLineWithSkippedArgs(c, 0);
				break;
			case 'C':
				handleLineWithSkippedArgs(c, 4);
				break;
			case 'S':
			case 'Q':
				handleLineWithSkippedArgs(c, 2);
				break;
			case 'A':
				// Arc
				parsingPlanet = true;
				do {
					// The first five arguments can be ignored for our purposes
					for (int i=0; i<5; i++) getNumber();
					// The second two are the endpoints of the arc. They contribute towards the average
					// which determines the planet centre. Crude, but reasonably effective for the limited
					// purposes of this parser.
					auto x = getXCoord();
					auto y = getYCoord();
					planetCoordsTotalX += x;
					planetCoordsTotalY += y;
					planetCoordsCount++;
				} while(numberIsNext());
				break;
			default:
				break;
		}
	}
	
	endRift();
	endPlanet();
	
	return iniData;
}

char[] global_iniData;
Planet[] global_planets;
Rift[] global_rifts;

/// Call parsePathData on the values of the "d" attributes of all SVG <path> elements from the input, in order.
void collectPathData(XmlNode node, Matrix transform) {
	foreach (char[] childName, XmlNode[] childrenOfThatName; node.children()) {
		if (childName == "path") {
			foreach (XmlNode child; childrenOfThatName) {
				if (child.hasAttribute("d")) {
					// Apply transform if necessary
					auto newTransform = transform;
					if (child.hasAttribute("transform")) {
						newTransform = transform * Matrix.fromSVG(child.attribute("transform"));
					}
					// Get ID and Inkscape's custom (non-SVG-standard) label attribute
					// For rifts, the ID and label both show up in comments.
					// For planets, the ID shows up in a comment, and the label is the planet name.
					char[] id = child.hasAttribute("id") ? child.attribute("id") : "";
					char[] label = child.hasAttribute("inkscape:label") ? child.attribute("inkscape:label") : "";
					// Parse this path data
					global_iniData ~= parsePathData(child.attribute("d"), newTransform, id, label);
				}
			}
		} else {
			foreach (XmlNode child; childrenOfThatName) {
				// Apply transform if necessary
				auto newTransform = transform;
				if (child.hasAttribute("transform")) {
					newTransform = transform * Matrix.fromSVG(child.attribute("transform"));
				}
				// Recurse
				collectPathData(child, newTransform);
			}
		}
	}
}


char[] parse(char[] filename, XmlNode doc) {
	char[] iniFile = "[map]\nname="~getName(getBaseName(filename))~"\nauthor=\ndescription=\n";
	
	global_planets.length = 0;
	global_rifts.length = 0;
	global_iniData = "";
	
	collectPathData(doc, Matrix.makeIdentity());
	iniFile ~= global_iniData;
	
	return iniFile;
}


unittest {
	scaleFactor = 1.333108/1.5;
	scope(exit) scaleFactor = 1;
	char[] testfilename = "svgtomap_unittest.svg";
	char[] testSVG="
<svg>
	<g>
		<!-- Closed arcs like this one are interpreted as planets -->
		<path d=\"M 171.42858 214.50504 A 7.1428571 6.4285712 0 1 1  157.14286,214.50504 A 7.1428571 6.4285712 0 1 1  171.42858 214.50504 z\" inkscape:label=\"Planet Name\" />
		<!-- Sequences of straight line drawing commands are interpreted as rifts -->
		<g transform=\"\">
			<path d=\"M 484.28571,338.07647 L 612.85714,342.36218 L 700,342.36218\" id=\"nestedpath\"/>
		</g>
		<path transform=\"translate(100,0) translate(0,10)\" d=\"M 300 328.07647 L 512.85714 332.36218 600 332.36218\" />
	</g>
</svg>";
	
	std.file.write(testfilename, testSVG);
	scope auto doc = new XmlNode(new File(testfilename));
	std.file.remove(testfilename);
	
	/* Expected output if each path becomes one rift
	char[] expectedOutput = "[map]
name=svgtomap_unittest
author=
description=

[planet]
name=Planet Name
x=219
y=286

[rift]
x1=533
y1=451

x2=817
y2=456

x3=933
y3=456

[rift]
; ID: nestedpath
x1=646
y1=451

x2=817
y2=456

x3=933
y3=456
";*/
	char[] expectedOutput = "[map]
name=svgtomap_unittest
author=
description=

[planet]
name=Planet Name
x=219
y=286

[rift]
x1=533
y1=451

x2=817
y2=456

[rift]
x1=817
y1=456

x2=933
y2=456

[rift]
; ID: nestedpath
x1=646
y1=451

x2=817
y2=456

[rift]
; ID: nestedpath
x1=817
y1=456

x2=933
y2=456
";
	char[] gotOutput = parse(testfilename, doc);
	assert(gotOutput == expectedOutput, "svgtomap unit test 1:\n-----\n"~gotOutput~"\n-----\n"~expectedOutput~"\n-----");
}


class Point2D {
	int x,y;
	
	this(int x, int y) {
		this.x = x;
		this.y = y;
	}
}


class Planet {
	Planet[] connected; /// List of planets directly reachable from this planet (no rift in the way)
	Point2D loc;
	bool flag = false;
	
	this(int x, int y) {
		loc = new Point2D(x, y);
	}
	
	char[] toString() {
		return "Planet at "~.toString(loc.x)~","~.toString(loc.y);
	}
}


class Rift {
	Point2D from, to;
	
	this(Point2D _from, Point2D _to) {
		from = _from;
		to = _to;
	}
	
	char[] toString() {
		return "Rift from "~.toString(from.x)~","~.toString(from.y)~" to "~.toString(to.x)~","~.toString(to.y);
	}
}


bool lineSegmentsIntersect(Point2D l1p1, Point2D l1p2, Point2D l2p1, Point2D l2p2) {
	// From "Algorithms in C" by Robert Sedgewick, 1990 edition, p.351
	return
		   ((ccw(l1p1, l1p2, l2p1)
		   * ccw(l1p1, l1p2, l2p2)) <= 0)
		&& ((ccw(l2p1, l2p2, l1p1)
		   * ccw(l2p1, l2p2, l1p2)) <= 0);
}

private int ccw(Point2D p0, Point2D p1, Point2D p2) {
	int dx1, dx2, dy1, dy2;
	dx1 = p1.x - p0.x; dy1 = p1.y - p0.y;
	dx2 = p2.x - p0.x; dy2 = p2.y - p0.y;
	if (dx1*dy2 > dy1*dx2) return +1;
	if (dx1*dy2 < dy1*dx2) return -1;
	if (dx1*dx2 < 0 || dy1*dy2 < 0) return -1;
	if ((dx1*dx1 + dy1*dy1) < (dx2*dx2 + dy2*dy2)) return +1;
	return 0;
}

unittest {
	assert(lineSegmentsIntersect(new Point2D(5,100), new Point2D(5,200) ,
	                             new Point2D(0,150), new Point2D(9,150) ));
	
	assert(lineSegmentsIntersect(new Point2D(5,101), new Point2D(5,200) ,
	                             new Point2D(0,150), new Point2D(9,151) ));
	
	assert(!lineSegmentsIntersect(new Point2D(5,101), new Point2D(5,200) ,
	                              new Point2D(0,100), new Point2D(4,151) ));
	
	assert(!lineSegmentsIntersect(new Point2D(5,100), new Point2D(5,200) ,
	                              new Point2D(0,150), new Point2D(4,150) ));
}


void flagRecursively(Planet start) {
	start.flag = true;
	foreach(Planet p; start.connected) {
		if (!p.flag) flagRecursively(p);
	}
}


bool fullyConnected(Planet[] planets, Rift[] rifts) {
	assert(planets.length);
	
	// Build connections
	for (int i=0; i<planets.length; i++) {
		for (int j=0; j<i; j++) {
			bool accessible = true;
			foreach (Rift rift; rifts) {
				if (lineSegmentsIntersect(rift.from, rift.to, planets[i].loc, planets[j].loc)) {
					accessible = false;
					break;
				}
			}
			if (accessible) {
				planets[i].connected ~= planets[j];
				planets[j].connected ~= planets[i];
			}
		}
	}
	
	// Walk the graph, flagging all connected planets
	flagRecursively(planets[0]);
	// Any unflagged planets?
	foreach (Planet p; planets) {
		if (!p.flag) return false;
	}
	return true;
}


int main(char[][] args) {
	if (args.length < 2 || args.length > 3) {
		writefln("Usage: svgtomap inputfile.svg [outputfile.ini]");
		return 1;
	}
	
	char[] inputFilename = args[1];
	char[] outputFilename = (args.length>=3 ? args[2] : addExt(args[1], "ini"));
	
	//writefln("Parsing %s, writing to %s...", inputFilename, outputFilename);
	
	// Input
	writefln("Reading %s...", inputFilename);
	Stream file = new BufferedFile(inputFilename);
	scope auto inputDocument = new XmlNode(file);
	// Parsing
	writefln("Parsing...");
	char[] iniFile = parse(inputFilename, inputDocument);
	
	//writefln("PRODUCED INI FILE:\n++++\n%s\n++++", iniFile);
	
	// Checking
	if (!global_planets.length) {
		writefln("ERROR: No planets found in SVG file. (This program assumes that subpaths containing SVG arcs are planets.)");
		return 1;
	} else if (!fullyConnected(global_planets, global_rifts)) {
		writefln("WARNING: Some parts of the map are completely cut off from other parts, making");
		writefln("         them inaccessible. The map will most likely not be playable.");
	}
	// Output
	writefln("Writing %s...", outputFilename);
	std.file.write(outputFilename, iniFile);
	
	writefln("All done!");
	return 0;
}

