
/**
 *	@file		XmlEntities.d
 *	@version	v1.06
 *	@date		14.10.2004
 * @author	Miguel Ferreira Simoes
 *        
 *	This software is provided 'as-is', without any express or implied
 *	warranty. In no event will the authors be held liable for damages
 *	of any kind arising from the use of this software.
 *      
 *	Permission is hereby granted to anyone to use this software for any 
 *	purpose, including commercial applications, and to alter it and/or 
 *	redistribute it freely, subject to the following restrictions:
 *       
 *	1. The origin of this software must not be misrepresented; you must 
 *		not claim that you wrote the original software.
 *	2. Altered source versions must be plainly marked as such, and must 
 *		not be misrepresented as being the original software.
 *	3.	This notice may not be removed or altered from any distribution
 *		of the source.
 *
 */

module xml.XmlEntities;

private import std.string;

/* External C library functions */

extern (C) {
	char *strpbrk(char *, char *);
}

/**
 * @class XmlEntities
 * This class is supposed to contain all xml built-in entities.
 *
 * Entities in principle are similar to simple C macros. 
 * An entity defines an abbreviation for a given string that you can
 * reuse many times throughout the content of your document.
 *
 * Entities can be encoded or decoded.
 */

class XmlEntities {

	/* Private Data */

	/**
	 * @struct Pair
	 * Internal representation of the entity.
	 * @var _encoded Encoded representation of the entity;
	 * @var _decoded Decoded representation of the entity;
	 */

	private struct Pair {
		char[] _encoded;
		char _decoded;
	}

	/// Contains the set of all xml built-in entities.

	private static const Pair[] _entities = [	{ "&amp;"	, '&'	},
																	{ "&lt;" 		, '<'	},
																	{ "&gt;" 		, '>'	},
																	{ "&apos;" 	, '\''	},
																	{ "&quot;" 	, '\"'	} ];

	/// String containing the union of all _decoded fields.
	
	private static char[] _decodedCharset;

	/* Constructors */

	/**
	 * Static Constructor.
	 * Initializes _decodedCharset static variable.
	 */

	public static this () {
		foreach(Pair pair; _entities)
			_decodedCharset ~= pair._decoded;
	}

	/* Public Methods */

	/**
	 * Static method that replaces all occurrences of decoded entities.
	 * @param src Source string containing decoded data.
	 * @return Encoded data string.
	 */

	public static char[] encode (char[] src) {
		char* ptr;
		char[] tmp;
		uint index, indexFind;
		
		for (index = 0; index < src.length; index = indexFind + 1) {

			if ((ptr = strpbrk(src[index .. src.length].ptr, _decodedCharset.ptr)) == null) {
				tmp ~= src[index .. src.length];
				break;
			}
			indexFind = index + ptr - cast(char *)src[index .. src.length];
			tmp ~= src[index .. indexFind];
			
			bit stop = false;
			foreach(Pair pair; _entities) {
				if (src[indexFind] == pair._decoded) {
					tmp ~= pair._encoded;
					stop = true;
					break;
				}
			}
			if (stop == false)
				tmp ~= src[indexFind];	
		}
		return tmp;
	}
	
	/**
	 * Static method that replaces all occurrences of encoded entities.
	 * @param src Source string containing encoded data.
	 * @return Decoded data string.
	 */
	
	public static char[] decode (char[] src) {
		char[] tmp;
		uint index, indexFind, indexPair;
				
		for (index = 0; index < src.length; ) {
			indexFind = find(src[index .. src.length], "&");
			if (indexFind == -1) {
				tmp ~= src[index .. src.length];
				break;
			}
			tmp ~= src[index .. index + indexFind];
			index += indexFind;
	
			bit stop = false;
			foreach(Pair pair; _entities) {
				
				if (index + pair._encoded.length <= src.length && 
					src[index + 1 .. index + pair._encoded.length] == pair._encoded[1 .. length]) {
						
					tmp ~= pair._decoded;
					index += pair._encoded.length;
					stop = true;
					break;
				}
			}
			if (stop == false) {
				tmp ~= '&';
				index += 1;
			}
		}
		return tmp;
	}

}

unittest {

	char[] encoded = "sep1&amp;sep2&lt;sep3&gt;sep4&apos;sep5&quot;sep6";
	char[] decoded = "sep1&sep2<sep3>sep4\'sep5\"sep6";

	assert(XmlEntities.encode(decoded) == encoded);
	assert(XmlEntities.decode(encoded) == decoded);
		
}
