
/**
 *	@file		XmlNode.d
 *	@version	v1.06
 *	@date		14.10.2004
 * @author	Miguel Ferreira Simoes
 *        
 *	This software is provided 'as-is', without any express or implied
 *	warranty. In no event will the authors be held liable for damages
 *	of any kind arising from the use of this software.
 *      
 *	Permission is hereby granted to anyone to use this software for any 
 *	purpose, including commercial applications, and to alter it and/or 
 *	redistribute it freely, subject to the following restrictions:
 *       
 *	1. The origin of this software must not be misrepresented; you must 
 *		not claim that you wrote the original software.
 *	2. Altered source versions must be plainly marked as such, and must 
 *		not be misrepresented as being the original software.
 *	3.	This notice may not be removed or altered from any distribution
 *		of the source.
 *
 */

module xml.XmlNode;

private import std.ctype;
private import std.stream, std.file;

private import xml.XmlStream;
private import xml.XmlEntities;
private import xml.XmlException;


/**
 * @class XmlNode
 * Represents a xml node. In fact this is the project main class.
 * A node has a name, and may have data (value), attributes or nodes.
 *
 * It is not supposed to implement w3c Document Object Model (DOM).
 * It is just a simple, fast and easy to learn xml parser. 
 */

class XmlNode {

	/* Private data */
	
	private char[] _name;
	private char[] _value;
	private char[][char[]] _attributes;
	private XmlNode[][char[]] _children;

	/**
	 * Constructor.
	 * @param name The name of the node.  
	 */

	public this (char[] name) {
        _name = name;
    }

	/* Constructors */

	/**
	 * Constructor.
	 * @param stream Stream containing a xml file or fragment.
	 */

	public this (Stream stream) {
		XmlStream xmlstream = new XmlStream(stream);
	
		xmlstream.read();
		if (xmlstream.skipUnparsedElements() == false)
			throw new XmlException(XmlException.Type.formed, xmlstream.readLine());
		this(xmlstream);
	}

	/**
	 * Private Constructor. Internal Usage only.
	 * It is used recursively by itself to append a new node children.
	 * @param stream XmlStream containing xml data.
	 */	

	private this (XmlStream stream) { 

		if (readElement(stream) == false)
			return;
		
		for(bit stop = false; stop == false;) {	
			if (stream.skipUnparsedElements() == true) {
				switch(stream.readCurrent()) {
					case '/':	stream.read();
									if (_name == stream.readName() && stream.skipWhiteSpaceCurrent() != '>') 
										throw new XmlException(XmlException.Type.endTag, stream.readLine());
									
									stop = true;
									break;

					default:	children(new XmlNode(stream));
				}
			}
			else {
				for(bit wspace = true; stream.read() != '<';) {
					if (find(whitespace, stream.readCurrent()) == -1) {
						_value ~= stream.readCurrent(); 
						wspace = false;
					}
					else if (!wspace) {
						_value ~= ' ';
						wspace = true;
					}
				}
				
				if (_value.length) {
					if (_value[length - 1] == ' ')
						_value.length = _value.length - 1;
					_value = XmlEntities.decode(_value);	
				}	
			}
			
		}
	}

	/* Public Acessor Methods */

	public char[] name () {
		return _name;
	}
	
	public char[] value () {
		return _value;
	}
	public XmlNode value (char[] value) {
		_value = XmlEntities.decode(value);
		return this;
	}
	
	public char[][char[]] attribute () {
		return _attributes;
	}
	public char[] attribute (char[] name) {
		return _attributes[name];
	}
	public XmlNode attribute (char[] name, char[] value) {
		_attributes[name] = XmlEntities.decode(value);
		return this;
	}
	
	public XmlNode[][char[]] children () {
		return _children;
	}
	public XmlNode[] children (char[] name) {
		return _children[name];
	}
    public XmlNode children (XmlNode newNode) {
		_children[newNode.name()] ~= newNode;
	    return this;
    }
	
	/* Public Boolean Query Methods */

	public bit hasValue () {
		return _value.length != 0;
	}

	public bit hasChildren() {
		return _children.length != 0;
	}
	public bit hasChildren (char[] name) {
		return (name in _children) !is null;
	}
		
	public bit hasAttribute () {
		return _attributes.length != 0;
	}
	public bit hasAttribute (char[] name) {
		return (name in _attributes) !is null;
	}
	
	/* Write Method */
		
	/**
	 * Writes xml node to a stream.
	 * @param stream Destiny stream. XmlNode is written to stream.
	 * @param indent Indentation expressed in whitespaces units.
	 *							 If indentation is 0: xml declaration is appended.
	 * @param tabsize Tabsize in whitespaces units.
	 */	
		
	public void write (Stream stream, uint indent = 0, uint tabsize = 4) { 
	
		char[] tabs = new char[indent + tabsize];
		tabs[] = ' ';
		
		if (indent == 0)
			stream.writeLine(tabs[0 .. indent] ~ "<?xml version=\"1.0\" encoding=\"ascii\"?>");
		
		char[] tag = tabs[0 .. indent] ~ "<" ~ _name;

		foreach(char[] name, char[] value; _attributes)
			tag ~= " " ~ name ~ "=\"" ~ XmlEntities.encode(value) ~ "\""; 

		if (_children.length == 0 && _value.length == 0) 
			stream.writeLine(tag ~= "/>");
		
		else { 
			stream.writeLine(tag ~= ">");
			
			foreach(XmlNode[] nodeArray; _children)
				foreach(XmlNode node; nodeArray)
					node.write(stream, tabs.length , tabsize);

			if (_value.length > 0)
				stream.writeLine(tabs ~ XmlEntities.encode(_value));

			stream.writeLine(tabs[0 .. indent] ~ "</" ~ _name ~ ">");
		}
	}

	/* Private Methods */

	/**
	 * Parses an element open tag stored in a stream.
	 * Node name and attributes are filled.
	 * @param stream XmlStream containing the Stream with xml data.
	 * @return true if it is an open only tag. false is the tag opens and closes the node.
	 */	

	private bit readElement (XmlStream stream) {
			
		for(_name = stream.readName(); true; stream.read()) {
	
			if (find(whitespace, stream.readCurrent()) != -1)
				stream.skipWhiteSpace();
				
			else if (find("//>", stream.readCurrent()) == -1)
				throw new XmlException(XmlException.Type.name, stream.readLine());

			switch(stream.readCurrent()) {
				case '>':	return true;
				case '/':	if (stream.read() == '>')
									return false;
								else
									throw new XmlException(XmlException.Type.name, stream.readLine());
				default:
			}
			
			char[] attributeName = stream.readName();
			
			if (stream.skipWhiteSpaceCurrent() != '=' || stream.skipWhiteSpace() != '\"')
				throw new XmlException(XmlException.Type.attributes, stream.readLine());
					
			char[] attributeValue;
			attributeValue.length = 0;
		
			while(stream.read() != '\"')
				attributeValue ~= stream.readCurrent();

			attribute(attributeName, attributeValue);
		}
	}

}


unittest {

	void checkAssertions(XmlNode root) {
		assert(root.name == "root");
		assert(root.value == "root data&end");
		assert(root.attribute("a1") == "root a1");
		assert(root.attribute("a2") == "root a2");

		assert(root.hasValue == true);
		assert(root.hasChildren == true);
		assert(root.hasChildren("son1") == true);
		assert(root.hasChildren("son4") == false);
		assert(root.hasAttribute == true);
		assert(root.hasAttribute("a1") == true);
		assert(root.hasAttribute("a3") == false);
	
		assert(root.children("son2")[0].hasValue == false);
		assert(root.children("son2")[0].hasChildren == false);
		assert(root.children("son2")[0].hasAttribute == false);
	
		assert(root.children("son3")[0].name == "son3");
		assert(root.children("son3")[0].value == "son3 data");
		assert(root.children("son3")[0].attribute("a1") == "son3 a1");
	}
	
	XmlNode root = new XmlNode("root");
	root.value("root data&end");
	root.attribute("a1", "root a1");
	root.attribute("a2", "root a2");
	
	root.children((new XmlNode("son1")).children(new XmlNode("son1-son1")));
	root.children(new XmlNode("son2"));
	root.children((new XmlNode("son3"))
		.value("son3 data")
		.attribute("a1", "son3 a1")
		.children(new XmlNode("son3-son1"))
	);
	checkAssertions(root);

	BufferedFile file1 = new BufferedFile("zXml.xml", FileMode.OutNew);
	root.write(file1);
	file1.close();

	BufferedFile file2 = new BufferedFile("zXml.xml", FileMode.In);
	XmlNode rootFile = new XmlNode(file2);
	//rootFile.write(std.stream.stdout);
	checkAssertions(rootFile);
	file2.close();
	std.file.remove("zXml.xml");
}
