
/**
 *	@file		XmlStream.d
 *	@version	v1.06
 *	@date		14.10.2004
 * @author	Miguel Ferreira Simoes
 *        
 *	This software is provided 'as-is', without any express or implied
 *	warranty. In no event will the authors be held liable for damages
 *	of any kind arising from the use of this software.
 *      
 *	Permission is hereby granted to anyone to use this software for any 
 *	purpose, including commercial applications, and to alter it and/or 
 *	redistribute it freely, subject to the following restrictions:
 *       
 *	1. The origin of this software must not be misrepresented; you must 
 *		not claim that you wrote the original software.
 *	2. Altered source versions must be plainly marked as such, and must 
 *		not be misrepresented as being the original software.
 *	3.	This notice may not be removed or altered from any distribution
 *		of the source.
 *
 */

module xml.XmlStream;

private import std.ctype;
private import std.stream;
private import xml.XmlException;

/**
 * @class XmlStream
 * This class is supposed to wrap a Stream instance.
 *
 * All read methods use Stream.read() method instead of 
 * Stream.getc()/ungetc() because it is faster.
 *
 * The last character is stored in a variable(char _current);
 *
 * Additional functionalities are available in order to:
 * 	- skip whitespaces
 * 	- skip unparsed elements types (comments, doctype and processingInstruction
 */

class XmlStream {
	
	/* Private Data */
	
	private uint _line;
	private char _current;
	private Stream _stream;
	private static char[] _nameBegin;
	private static char[] _nameBody;

	/* Constructor */

	/**
	 * Static Constructor.
	 * Initializes static variables:
	 *		- _nameBegin: all chars that can start a xml name.
	 *		- _nameBody : all chars that can be in the body of a xml name.
	 */

	public static this() {
		_nameBegin = letters ~ "_";
		_nameBody = letters ~ digits ~ "_-.:";
	}

	/**
	 * Constructor.
	 * Wraps a Stream instance.
	 * Initializes line number with 0.
	 */

	public this(Stream stream) {
        _line = 0;
		_current = 0;
        _stream = stream;
    }

	/* Public Methods */
	
	/// Reads and returns a character from the stream updating current
	
	public char read() {
		_stream.readExact(&_current, _current.sizeof);
		if (_current == '\n')
			_line++;
		return _current;
	}

	/// Returns the line number

	public uint readLine() {
		return _line;
	}

	/// Returns the last character

	public char readCurrent() {
		return _current;
	}
	
	/// Reads a string with the given size
	
	public char[] readString(uint size) {
		char[] tmp;
		for (uint i = 0; i < size; ++i)
			tmp ~= read();
		return tmp;
	}

	/// Reads a valid xml name
	
	public char[] readName () {
		char[] name = std.string.toString(_current);
		
		if (find(_nameBegin, _current) == -1)
			throw new XmlException(XmlException.Type.name, _line);

		while(find(_nameBody, read()) != -1)
			name ~= _current; 
		return name;
	}

	/// Reads the stream and returns the first non-whitespace char

	public char skipWhiteSpace () {
		while(find(whitespace, read()) != -1) { }
		return _current;
	}

	/// Returns first non-whitespace char including last char read (current)

	public char skipWhiteSpaceCurrent () {
		return find(whitespace, _current) == -1 ? _current : skipWhiteSpace();
	}
	
	/// Reads the stream and skips all non-parsed elements
	
	public bit skipUnparsedElements () {
		while(true) {
			if (skipWhiteSpaceCurrent() == '<') {
				switch (read()) {
					case '?':	skipProcessingInstruction();
									break;
					case '!':	char[] aux = readString(2);
									if (aux == "--")
										skipComment();
									else if (aux ~ readString(5) == "DOCTYPE")
										skipDocumentType();
									else
										throw new XmlException(XmlException.Type.unrecognizedTag, _line);
									break;	
					default:	return true;
				}
				read();
			}
			else return false;
		}
	}
	
	/* Private Methods */
	
	/// Skips the comment tag
	
	private void skipComment () {
		char[2] last;
		while (read() != '>' || last != "--") {
			last[0] = last[1];
			last[1] = _current;
		}
	}
	
	/// Skips the DOCTYPE tag
	
	private void skipDocumentType ()  {
		while(read() != '>') { }
	}
	
	/// Skips any processing instruction tag
		
	private void skipProcessingInstruction () {
		char last;
		for(; read() != '>'; last = _current) { }
		
		if (last != '?')
			throw new XmlException(XmlException.Type.processingInstruction, _line);
	}
	
}